# -*- coding: utf-8 -*-
# @Time    : 2020/2/3 下午1:40
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : metrics.py
# @Desc    : 自定义评估指标

from .core import *


class BaseMetric(metaclass=ABCMeta):
    def __init__(self):
        pass

    @abstractmethod
    def reset(self):
        raise NotImplementedError('自定义指标必须实现该方法')

    @abstractmethod
    def __call__(self, y_pred: Tensor, y_true: Tensor):
        raise NotImplementedError('自定义指标必须实现该方法')


class BinaryAccuracy(BaseMetric):

    def __init__(self):
        super(BinaryAccuracy, self).__init__()
        self.correct_count = 0
        self.total_count = 0

    def reset(self):
        self.correct_count = 0
        self.total_count = 0

    def __call__(self, y_pred: Tensor, y_true: Tensor):
        self.correct_count += y_pred.round().eq(y_true.view(-1, 1)).float().sum().item()
        self.total_count += len(y_pred)
        acc = float(self.correct_count) / float(self.total_count)
        return np.round(acc, 4)



# -*- coding: utf-8 -*-
# @Time    : 2020/1/30 下午2:33
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : data.py
# @Desc    : 表格数据所需的Dataset


from ..core import *


class WideDeepDataset(Dataset):
    """Wide&Deep模型所需的数据集结构

    Args:
        x_wide (np.ndarray): Wide模块的输入特征
        x_deep (np.ndarray): DeepDense模块的输入特征
        y (np.ndarray): 预测目标值
    """
    def __init__(self,
                 x_wide: np.ndarray,
                 x_deep: np.ndarray,
                 y: Optional[np.ndarray] = None):
        self.x_wide = x_wide
        self.x_deep = x_deep
        self.y = y

    def __getitem__(self, idx: int):
        x = Bunch(wide=self.x_wide[idx], deep=self.x_deep[idx])
        if self.y is not None:
            return x, self.y[idx]
        return x

    def __len__(self):
        return len(self.x_deep)



# -*- coding: utf-8 -*-
# @Time    : 2020/2/1 下午12:47
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : preprocessing.py
# @Desc    : Wide&Deep模型预处理

from ..core import *
from .utils import MultiColumnLabelEncoder


class BasePreprocessor:
    def __init__(self):
        pass

    def fit(self, x):
        pass

    def transform(self, x) -> np.ndarray:
        pass

    def fit_transform(self, x) -> np.ndarray:
        pass


class WidePreprocessor(BasePreprocessor):
    """Wide模块的预处理器

    WidePreprocessor对于类别型变量进行OneHot编码,作为模型Wide部分的输入

    Args:
        wide_cols (List): 包含类别型变量名称的列表

    Attributes:
        encoder (OneHotEncoder): sklearn中的独热编码器
    """

    def __init__(self, wide_cols: List[str]):
        super(WidePreprocessor, self).__init__()
        self.wide_cols = wide_cols
        self.encoder = OneHotEncoder(sparse=False)

    def fit(self, x: pd.DataFrame) -> BasePreprocessor:
        x_wide = x[self.wide_cols].copy()
        self.encoder.fit(x_wide[self.wide_cols])
        return self

    def transform(self, x: pd.DataFrame) -> np.ndarray:
        x_wide = x[self.wide_cols].copy()
        return self.encoder.transform(x_wide)

    def fit_transform(self, x: pd.DataFrame) -> np.ndarray:
        return self.fit(x).transform(x)


class DeepPreprocessor(BasePreprocessor):
    """DeepDense模块的预处理器

    DeepPreprocessor包含两个部分:
    (1) 对于类别型变量进行标签编码,编码采用了自定义的MultiColumnLabelEncoder
    (2) 对于连续型变量进行标准化

    Args:
        embed_col_dims (List): 包含有(列名,嵌入维度)的列表
        cont_cols (List): 包含连续型变量名称的列表

    Attributes:
        column_idx (Dict): 列名的索引字典
        embed_cols (List): 包含需要嵌入的变量名称的列表
        embed_input (List): 包含有(列名,类型数目,嵌入维度)的元组列表
        encoder (MultiColumnLabelEncoder): 自定义的多列标签编码器
        scaler (StandardScaler): sklearn中的标准化器
    """

    def __init__(self,
                 embed_col_dims: List[Tuple[str, int]] = None,
                 cont_cols: List[str] = None):
        super(DeepPreprocessor, self).__init__()
        self.embed_col_dims = embed_col_dims
        self.cont_cols = cont_cols
        self.column_idx = dict()

        # 处理类别型特征时的属性
        self.embed_dim = None
        self.embed_cols = None
        self.embed_input = None
        self.encoder = None

        # 处理连续型特征时的属性
        self.scaler = None

    def fit(self, x: pd.DataFrame) -> BasePreprocessor:
        # 处理类别型变量
        if self.embed_col_dims is not None:
            self.embed_dim = dict(self.embed_col_dims)
            self.embed_cols = [e[0] for e in self.embed_col_dims]
            self.encoder = MultiColumnLabelEncoder(self.embed_cols)

            self.encoder.fit(x)
            self.embed_input = []
            for k, v in self.encoder.encoders.items():
                self.embed_input.append((k, len(v.classes_), self.embed_dim[k]))

        if self.cont_cols is not None:
            x_cont = x[self.cont_cols].copy()
            self.scaler = StandardScaler().fit(x_cont)

        return self

    def transform(self, x: pd.DataFrame) -> np.ndarray:
        x_embed, x_cont = None, None
        if self.embed_col_dims is not None:
            x_embed = self.encoder.transform(x)
        if self.cont_cols is not None:
            x_cont = x[self.cont_cols].copy().values
            x_cont = self.scaler.transform(x_cont)

        if x_embed is not None and x_cont is not None:
            x_out = np.concatenate([x_embed, x_cont], axis=1)
            columns = self.embed_cols + self.cont_cols
        elif x_embed is not None:
            x_out = x_embed
            columns = self.embed_cols
        else:
            x_out = x_cont
            columns = self.cont_cols

        self.column_idx = {v: k for k, v in enumerate(columns)}
        return x_out

    def fit_transform(self, x: pd.DataFrame) -> np.ndarray:
        return self.fit(x).transform(x)



# -*- coding: utf-8 -*-
# @Time    : 2020/2/2 下午1:45
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : learner.py
# @Desc    : Wide&Deep模型训练


from ..core import *
from ..metrics import BinaryAccuracy
from .data import WideDeepDataset


class WideDeepLearner:
    """Wide&Deep模型训练

    WideDeepLearner类对于Wide&Deep模型进行训练,可以解决回归问题和二分类问题

    Args:
        model (nn.Module): Wide&Deep模型
        objective (str): 训练目标(‘regression’ or 'binary')
    """

    def __init__(self, model: nn.Module, objective: str):
        self.model = model
        self.objective = objective
        self.optimizer = optim.Adam(self.model.parameters())

        # 设定评测指标,二分类问题默认采用准确率
        if self.objective == 'binary':
            self.metric = BinaryAccuracy()
        else:
            self.metric = None
        # 设定默认batch_size大小
        self.batch_size = 128

        if USE_CUDA:
            self.model.cuda()

        self._temp_train_loss = 0.
        self._temp_valid_loss = 0.

    def _split_dataset(self, x_wide: np.ndarray, x_deep: np.ndarray, y: np.ndarray, test_size: float = 0.2):
        """将数据划分为训练集和验证集

        Args:
            x_wide (np.ndarray): Wide模块输入
            x_deep (np.ndarray): DeepDense模块输入
            y (np.ndarray): 预测目标值
            test_size (float): 验证集比例,为0-1之间的浮点数

        Returns:
            train_dataset (WideDeepDataset): 训练集的Dataset
            valid_dataset (WideDeepDataset): 验证集的Dataset
        """
        x_train_wide, x_valid_wide, x_train_deep, x_valid_deep, y_train, y_valid = train_test_split(
            x_wide, x_deep, y, test_size=test_size,
            stratify=y if self.objective != 'regression' else None
        )
        x_train = {'x_wide': x_train_wide, 'x_deep': x_train_deep, 'y': y_train}
        x_valid = {'x_wide': x_valid_wide, 'x_deep': x_valid_deep, 'y': y_valid}

        train_dataset = WideDeepDataset(**x_train)
        valid_dataset = WideDeepDataset(**x_valid)
        return train_dataset, valid_dataset

    def _acti_func(self, x: Tensor) -> Tensor:
        if self.objective == 'binary':
            return torch.sigmoid(x)
        else:
            return x

    def _loss_func(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        if self.objective == 'binary':
            return F.binary_cross_entropy(y_pred, y_true.view(-1, 1))
        else:
            return F.mse_loss(y_pred, y_true.view(-1, 1))

    def _make_train_step(self, x: Dict[str, Tensor], y: Tensor, batch_idx: int):
        self.model.train()

        x = {k: v.cuda() for k, v in x.items()} if USE_CUDA else x
        y = y.float()
        y = y.cuda() if USE_CUDA else y

        self.optimizer.zero_grad()
        y_pred = self._acti_func(self.model(x))

        loss = self._loss_func(y_pred, y)
        loss.backward()
        self.optimizer.step()

        self._temp_train_loss += loss.item()
        train_loss = self._temp_train_loss / (batch_idx + 1)

        if self.metric:
            train_metric = self.metric(y_pred, y)
            return train_metric, train_loss
        return None, train_loss

    def _make_valid_step(self, x: Dict[str, Tensor], y: Tensor, batch_idx: int):
        self.model.eval()

        with torch.no_grad():
            x = {k: v.cuda() for k, v in x.items()} if USE_CUDA else x
            y = y.float()
            y = y.cuda() if USE_CUDA else y

            y_pred = self._acti_func(self.model(x))
            loss = self._loss_func(y_pred, y)
            self._temp_valid_loss += loss.item()
            valid_loss = self._temp_valid_loss / (batch_idx + 1)

        if self.metric:
            valid_metric = self.metric(y_pred, y)
            return valid_metric, valid_loss
        return None, valid_loss

    def fit(self,
            x_wide: np.ndarray,
            x_deep: np.ndarray,
            y: np.ndarray,
            num_epochs: int,
            batch_size: int):
        """根据指定的训练轮数和批大小训练模型

        Args:
            x_wide (np.ndarray): Wide模块输入
            x_deep (np.ndarray): DeepDense模块输入
            y (np.ndarray): 预测目标值
            num_epochs (int): 训练轮数
            batch_size (int): 每个Batch的大小
        """
        self.batch_size = batch_size
        train_dataset, valid_dataset = self._split_dataset(x_wide, x_deep, y)
        train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True, num_workers=CPU_COUNT)
        valid_loader = DataLoader(dataset=valid_dataset, batch_size=batch_size, shuffle=False, num_workers=CPU_COUNT)

        for epoch in range(num_epochs):
            # 模型训练
            if self.metric:
                self.metric.reset()
            self._temp_train_loss = 0.

            pbar = tqdm(train_loader)
            for batch_idx, (inputs, targets) in enumerate(pbar):
                pbar.set_description('Epoch %i' % (epoch + 1))
                train_metric, train_loss = self._make_train_step(inputs, targets, batch_idx)
                if train_metric:
                    pbar.set_postfix(metric=train_metric, loss=train_loss)
                else:
                    pbar.set_postfix(loss=train_loss)

            # 模型验证
            if self.metric:
                self.metric.reset()
            self._temp_valid_loss = 0.

            pbar = tqdm(valid_loader)
            for batch_idx, (inputs, targets) in enumerate(pbar):
                pbar.set_description('Validation')
                valid_metric, valid_loss = self._make_valid_step(inputs, targets, batch_idx)
                if valid_metric:
                    pbar.set_postfix(metric=valid_metric, loss=valid_loss)
                else:
                    pbar.set_postfix(loss=valid_loss)

    def predict(self, x_wide: np.ndarray, x_deep: np.ndarray) -> np.ndarray:
        x_test = {'x_wide': x_wide, 'x_deep': x_deep}
        test_dataset = WideDeepDataset(**x_test)
        test_loader = DataLoader(dataset=test_dataset,
                                 batch_size=self.batch_size,
                                 num_workers=CPU_COUNT,
                                 shuffle=False)

        self.model.eval()
        batch_pred_list = []
        with torch.no_grad():
            pbar = tqdm(test_loader)
            for inputs in pbar:
                pbar.set_description('Predict')
                x = {k: v.cuda() for k, v in inputs.items()} if USE_CUDA else inputs
                batch_pred = self._acti_func(self.model(x))
                batch_pred = batch_pred.cpu().data.numpy()
                batch_pred_list.append(batch_pred)
        self.model.train()

        if self.objective == 'regression':
            return np.vstack(batch_pred_list).squeeze(1)
        else:
            y_pred = np.vstack(batch_pred_list).squeeze(1)
            return (y_pred > 0.5).astype('int')



# -*- coding: utf-8 -*-
# @Time    : 2020/1/29 下午4:23
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : models.py
# @Desc    : Wide&Deep模型文件

from ..core import *


class Wide(nn.Module):
    """Wide&Deep模型Wide模块

    Wide类将独热编码后的数据通过简单的线性层进行处理

    Args:
        in_dim (int): Wide模块的输入维度
        out_dim (int): Wide模块的输出维度

    Attributes:
        linear (nn.Module): Wide模块的线性层
    """
    def __init__(self, in_dim: int, out_dim: int = 1):
        super(Wide, self).__init__()
        self.linear = nn.Linear(in_dim, out_dim)

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(x.float())


class DeepDense(nn.Module):
    """Wide&Deep模型DeepDense模块

    DeepDense类对类别型变量进行嵌入表示并与连续型变量进行拼接
    拼接后的结果通过一系列的隐藏层输出

    Args:
        column_idx (Dict): 列名索引字典,用于将张量切片
        hidden_nodes (List): 隐藏层节点数量
        hidden_drop_ps (List): 隐藏层Dropout几率
        batch_norm (bool): 隐藏层是否使用Batch Normalization
        embed_input (List): 包含有(列名,类型数目,嵌入维度)的元组列表
        embed_drop_p (float): 嵌入层的Dropout几率
        cont_cols (List): 连续型变量的列名

    Attributes:
        embed_layers (nn.ModuleDict): 嵌入层模块字典
        dense_layers (nn.Sequential): 接受嵌入层输出和连续型变量的隐含层
        out_dim (int): DeepDense模块输出维度
    """
    def __init__(self,
                 column_idx: Dict[str, int],
                 hidden_nodes: List[int],
                 hidden_drop_ps: Optional[List[float]] = None,
                 batch_norm: bool = False,
                 embed_input: Optional[List[Tuple[str, int, int]]] = None,
                 embed_drop_p: float = 0.,
                 cont_cols: Optional[List[str]] = None):
        super(DeepDense, self).__init__()
        self.column_idx = column_idx
        self.batch_norm = batch_norm
        self.embed_input = embed_input
        self.cont_cols = cont_cols

        # 类别型变量
        if self.embed_input is not None:
            self.embed_layers = nn.ModuleDict({
                'embed_layer_' + col: nn.Embedding(num, dim)
                for col, num, dim in self.embed_input
            })
            self.embed_dropout = nn.Dropout(embed_drop_p)
            embed_in_dim = np.sum([x[2] for x in self.embed_input])
        else:
            embed_in_dim = 0

        # 连续型变量
        cont_in_dim = len(self.cont_cols) if self.cont_cols is not None else 0

        # 构建隐藏层
        in_dim = embed_in_dim + cont_in_dim
        hidden_nodes = [in_dim] + hidden_nodes
        if not hidden_drop_ps:
            hidden_drop_ps = [0.] * (len(hidden_nodes) - 1)
        self.dense_layers = nn.Sequential()
        for i in range(1, len(hidden_nodes)):
            self.dense_layers.add_module(
                'dense_layer_{}'.format(i - 1),
                self._create_dense_layer(
                    hidden_nodes[i - 1],
                    hidden_nodes[i],
                    hidden_drop_ps[i - 1],
                ))
        self.out_dim = hidden_nodes[-1]

    def _create_dense_layer(self,
                            in_dim: int,
                            out_dim: int,
                            drop_p: float = 0.):
        layers = [nn.Linear(in_dim, out_dim), nn.LeakyReLU(inplace=True)]
        if self.batch_norm:
            layers.append(nn.BatchNorm1d(out_dim))
        layers.append(nn.Dropout(drop_p))
        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x_emb, x_cont = None, None

        # 类别型变量
        if self.embed_input is not None:
            x_emb = [
                self.embed_layers['embed_layer_' + col](
                    x[:, self.column_idx[col]].long())
                for col, _, _ in self.embed_input
            ]
            x_emb = torch.cat(x_emb, 1)
            x_emb = self.embed_dropout(x_emb)
        # 连续型变量
        if self.cont_cols is not None:
            cont_idx = [self.column_idx[col] for col in self.cont_cols]
            x_cont = x[:, cont_idx].float()

        if x_emb is None:
            x_out = x_cont
        else:
            x_out = x_emb if x_cont is None else torch.cat([x_emb, x_cont], 1)
        return self.dense_layers(x_out)


class WideDeep(nn.Module):
    """Wide&Deep模型

    Google的Wide&Deep算法实现,用于处理传统表格数据
    论文参考: https://arxiv.org/abs/1606.07792
    实现参考: https://github.com/jrzaurin/pytorch-widedeep

    Args:
        wide (nn.Module): Wide&Deep模型中的Wide模块
        deepdense (nn.Module): Wide&Deep模型中的DeepDense模块
        out_dim (int): 最终输出的维度,'1'对应回归或者二分类问题,'n_class'对应多分类问题
    """
    def __init__(self, wide: nn.Module, deepdense: nn.Module,
                 out_dim: int = 1):
        super(WideDeep, self).__init__()
        self.wide = wide
        self.deepdense = nn.Sequential(deepdense,
                                       nn.Linear(deepdense.out_dim, out_dim))

    def forward(self, x: Dict[str, Tensor]) -> Tensor:
        """
        Args:
            x (Dict): 包含模型名('wide', 'deep')与对应张量的字典
        """
        x_out = self.wide(x['wide'])
        x_out.add_(self.deepdense(x['deep']))
        return x_out



# -*- coding: utf-8 -*-
# @Time    : 2020/2/1 下午2:26
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : utils.py
# @Desc    : 表格模型工具类

from ..core import *


def get_embed_col_dims(df: pd.DataFrame, cols: List[str]) -> List[Tuple[str, int]]:
    """获取默认配置下的嵌入维度

    Args:
        df (pd.DataFrame): 原始数据
        cols (List): 需要嵌入的列名

    Returns:
        embed_col_dims (List): 嵌入维度列表
    """
    embed_col_dims = []
    for c in cols:
        n_cat = len(df[c].unique())
        n_dim = min(600, round(1.6 * n_cat ** 0.56))
        embed_col_dims.append((c, n_dim))
    return embed_col_dims


class MultiColumnLabelEncoder:
    """多列标签编码器,用于神经网络Embedding的输入

    通过使用多个sklearn中的LabelEncoder对每一列进行编码

    Args:
        cols (List): 需要进行编码的列名

    Attributes:
        encoders (Dict): 列名与对应的LabelEncoder字典
    """

    def __init__(self, cols: List[str]):
        self.cols = cols
        self.encoders = {c: LabelEncoder() for c in cols}

    def fit(self, x: pd.DataFrame):
        for c in self.cols:
            self.encoders[c].fit(x[c])
        return self

    def transform(self, x: pd.DataFrame) -> np.ndarray:
        x_out = x[self.cols].copy()
        for c in self.cols:
            x_out[c] = self.encoders[c].transform(x_out[c])
        return x_out.values

    def fit_transform(self, x: pd.DataFrame) -> np.ndarray:
        return self.fit(x).transform(x)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

exec(open('regius/version.py').read())

setup(
    name='regius-dl',
    version=__version__,
    packages=find_packages(),

    description='Regius Deep Learning Toolbox',
    long_description=readme,
    long_description_content_type='text/markdown',

    author='Jiang Yize',
    author_email='315135833@qq.com',
    url='https://github.com/RegiusQuant/regius',

    license='Apache Software License 2.0',
    keywords=['deep learning', 'machine learning'],

    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: Chinese (Simplified)',
        'Programming Language :: Python :: 3.7',
    ],

    install_requires=[
        'pandas',
        'torch',
    ]
)



# -*- coding: utf-8 -*-
# @Time    : 2020/1/30 上午10:02
# @Author  : RegiusQuant <315135833@qq.com>
# @Project : regius
# @File    : core.py
# @Desc    : Regius常用类型和函数

import os
from abc import ABCMeta, abstractmethod

# typing相关导入
from typing import Tuple, List, Dict, Optional

import numpy as np
import pandas as pd

# sklearn相关导入
from sklearn.preprocessing import OneHotEncoder, LabelEncoder, StandardScaler
from sklearn.utils import Bunch
from sklearn.model_selection import train_test_split

# torch相关导入
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch import Tensor
from torch.utils.data import Dataset, DataLoader

from tqdm import tqdm, trange

CPU_COUNT = os.cpu_count()
USE_CUDA = torch.cuda.is_available()
